import warnings
from collections import defaultdict
import os
import re
import json
from posixpath import join as urljoin
from pathlib import Path
from math import floor
from functools import lru_cache

from fsspec import filesystem
from maxar_canvas_grid import covers, Cell
from maxar_canvas_grid.grid import get_transform
from max_ard.exceptions import MissingDependency

try:
    from shapely.geometry import mapping, shape, GeometryCollection
    from shapely.ops import transform
    HAS_SHAPELY = True
except ImportError:
    HAS_SHAPELY = False

try:
    import rasterio
    HAS_RASTERIO = True
except ImportError:
    HAS_RASTERIO = False

try:
    import fiona
    HAS_FIONA = True
except ImportError:
    HAS_FIONA = False

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property

catalog_url = 'https://geobigdata.io/catalog/v2/record/'

class ARDCollection():

    storage_cache = {}

    def __init__(self, path, aoi=None, acq_id_in=None, zone_in=None, earliest_date=None, latest_date=None):

        self.acq_bins = None
    
        # TODO we may want to normalize AOIs to WGS84
        # but right now `covers` is probably capable enough
        self.aoi = aoi

        assert zone_in is None or all(type(z) == int for z in zone_in), 'Zones must be integers'
        if not zone_in:
            zone_in = None
        self.zone_in = zone_in 

        assert acq_id_in is None or all(type(c) == str for c in acq_id_in), 'Catalog IDs must be strings'
        if not acq_id_in:
            acq_id_in = None
        self.acq_id_in = acq_id_in 

        def format_date(d):
            if d is None:
                return None
            try:
                return strftime('%Y-%m-%d')
            except:
                assert re.match(r'\d{4}-\d{2}-\d{2}', d), "Dates must be YYYY-MM-DD strings, or date/datetime objects"
                return d

        self.earliest_date = format_date(earliest_date)
        self.latest_date = format_date(latest_date)

        self.path = path
        #trailing slashes need to required or removed, try:
        # this might need to be smarter (os.path?) for windows slashes

        if self.path[-1] == '/':
            self.path = self.path[:-1]
        self.storage_type = 's3' 
        if os.path.exists(self.path):
            self.storage_type = 'file'
        if not self.path in ARDCollection.storage_cache:
            ARDCollection.storage_cache[self.path] = filesystem(self.storage_type)
        self.fs = ARDCollection.storage_cache[self.path]

    def __repr__(self):
        return f'<ARDCollection at {self.path}/>'

    # This is an optimization, need to proxy through with the ARD API
    # instead of using GBDX
    # @lru_cache()
    # def get_footprint(self, acq_id):
    #     try:
    #         r = gbdx.get(catalog_url + acq_id)
    #         r.raise_for_status()
    #         footprint = (r.json()['properties']['footprintWkt'])
    #         return wkt.loads(footprint) 
    #     except Exception as e:
    #         warnings.warn(e)

    def _ls(self):
        self._dates = set()
        self._zones = set()
        # boundary optimization for acquisition ID searches
        #if self.acq_id_in is not None:
            # grab the bounds from the catalog to speed up fetching
            # uses just zones for now, could quadkey too
          #  if self.zone_in is None and self.aoi is None:
          #      lon2zone = lambda lon: (floor((lon + 180)/6) % 60) + 1
                #  for acq_id in self.acq_id_in:
                #     try:
                #         xmin, ymin, xmax, ymax = self.get_footprint(acq_id).bounds 
                #         print(xmin, ymin, xmax, ymax)
                #         self._zones.add(lon2zone(xmin))
                #         self._zones.add(lon2zone(xmax))
                #     except:
                #         pass 
        if self.zone_in is not None:
            for zone in self.zone_in:
                self._zones.add(zone)
        self.acq_bins = defaultdict(list)
        qkbs = None
        # build leading glob pattern based on zone & quadkey
        if self.aoi is not None:
            cells = set([f'{c.zone}/{c.quadkey}' for c in covers(self.aoi)])
            qkbs = set()
            for qk in cells:
                qkbs.add(qk[:5])
            paths = [f'{k}*/*/*.json' for k in qkbs]
        else:
            if not self.zone_in or len(self.zone_in) == 0:
                paths =['*/*/*/*.json']
            else: 
                paths = [f'{z}/*/*/*.json' for z in self.zone_in]
        for path in paths:
            full_path = f'{self.path}/{path}'
            for item in self.fs.glob(full_path):
                tile = ARDTile.from_doc(self.fs, item)
                if qkbs is not None:
                    if f'{tile.zone}/{tile.quadkey}' not in cells:
                        continue
                if self.acq_id_in is not None and tile.acq_id not in self.acq_id_in:
                    continue
                if self.earliest_date is not None:
                    if tile.date < self.earliest_date:
                        continue
                if self.latest_date is not None:
                    if tile.date > self.latest_date:
                        continue
                self._dates.add(tile.date)
                self._zones.add(tile.cell.zone)
                self.acq_bins[tile.acq_id].append(tile)

    def clear_filesystem_cache(self):
        self.fs.clear_instance_cache()

   
    def update(func):
        ''' update decorator - tiles in collection should be updated before execution '''
        def update_wrapper(*args):
            self = args[0]
            if self.acq_bins is None:
                self._ls()
            return func(self)
        return update_wrapper

    @property
    @update
    def tiles(self):
        return [tile for tiles in self.acq_bins.values() for tile in tiles]

    @property
    @update
    def acq_ids(self):
        return list(self.acq_bins.keys())

    @property
    @update
    def acquisitions(self):
        return [Acquisition(acq_id, tiles, self) for acq_id, tiles in self.acq_bins.items()]

    @property
    @update
    def dates(self):
        dates = list(self._dates)
        dates.sort()
        return dates

    @property
    @update
    def start_date(self):
        return self.dates[0]

    @property
    @update
    def end_date(self):
        return self.dates[-1]

    @property
    @update
    def zones(self):
        return list(self._zones)

    def __setattr__(self, name, value):
        if name == 'acq_ids':
            raise ValueError('.acq_ids is read-only - set .acq_id_in instead')
        if name == 'zones':
            raise ValueError('.zones is read-only - set .zone_in instead')
        if name == 'start_date':
            raise ValueError('.start_date is read-only - set .earliest_date instead')
        if name == 'acq_ids':
            raise ValueError('.end_date is read-only - set .latest_date instead')
        if name in ['acq_id_in', 'zone_in', 'earliest_date', 'latest_date', 'aoi']:
            if self.acq_bins is not None:
                self.acq_bins = None
        object.__setattr__(self, name, value)

class Acquisition():
    ''' container for a bunch of tiles from a single acquisition '''

    def __init__(self, acq_id, tiles, parent):
        self.acq_id = acq_id
        self.tiles = tiles
        self.fs = parent.fs
        self.path = parent.path

    def __repr__(self):
        return f'<Acquisition of {self.acq_id} ({len(self.tiles)} tiles)>'

    @property
    def properties(self):
        ''' strip level props, maybe just fetch from DUC 
        
            Here we just pull common props from the first tile's props '''

        props = self.tiles[0].properties
        keep = ['catalog_id', 'platform', 'scan_direction','epsg','datetime']
        strip_props = {}
        for key in keep:
            strip_props[key] = props[key]
        return strip_props

    @property
    def stats(self):
        ''' Compute statistics about the tiles '''

        raise NotImplementedError
        stats = {
            'min_clouds': 101 ,
            'max_clouds': -1
        }
        # could multithread this too
        for tile in self.tiles:
            props = tile.properties
            stats['min_clouds'] = min(stats['min_clouds'], props['min_clouds'])
            # and so on, can clean up into function
        
        return stats
            

def copy(container, destination, flatten=False, dry_run=False):
    Path(destination).mkdir(parents=True, exist_ok=True)
    for tile in container.tiles:
        for path in tile.asset_paths.values():
            source = path
            target = path.replace(os.path.abspath(container.path), '')[1:]
            if flatten:
                target = urljoin(destination, target.replace('/','-'))
            else:
                target = urljoin(destination, target)
            if dry_run:
                print(source, target)
            else:
                if not flatten:
                    Path(target).parent.mkdir(parents=True, exist_ok=True)
                container.fs.get(source, target)


class ARDTile():
    # matching regexes
    # don't be a smartass and combine them

    # matches quadkeys
    CELL_REGEX = re.compile(r'[/-](\d{2})[-/]([0123]{12})[/-]')
    # matches regular catalog IDs
    CAT_ID_REGEX = re.compile(r'10[1-6][0AC][0-9A-F]{3}0[0-9A-F]{6}00', re.I)
    # matches WV4 catIDs (UUID v4)
    WV4_REGEX = re.compile(r'([0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}-inv)', re.I)
    # matches dates yyyy-mm-dd
    DATE_REGEX = re.compile(r'\d{4}-\d{2}-\d{2}')

    def __init__(self):
        self.date = None
        self.acq_id = None
        self.quadkey = None
        self.zone = None

    @cached_property
    def properties(self):
        return self.open_asset('item')['properties']

    @cached_property
    def clouds(self):
        try:
            return self.open_asset('clouds')
        except:
            return self.open_asset('_clouds')


    @cached_property
    def nodata(self):
        # TODO check projections
        return self.cell.geom_WGS84.difference(self.extent)

    @cached_property
    def extent(self):
        try:
            return self.open_asset('extent')
        except:
            return self.open_asset('_extent')
    
    @cached_property
    def cell(self):
        return Cell(self.quadkey, zone=self.zone)

    @classmethod
    def from_doc(cls, fs, path):

        tile = cls()
        tile.fs = fs

        # zone and quadkey
        matches = ARDTile.CELL_REGEX.search(path)
        if matches is not None:
            tile.zone, tile.quadkey = matches.groups()

        # cat ID
        matches = ARDTile.CAT_ID_REGEX.search(path)
        if matches is not None:
            tile.acq_id = matches[0]
        else:
            matches = ARDTile.WV4_REGEX.search(path)
            if matches is not None:
                tile.acq_id = matches[0]

        # date
        matches = ARDTile.DATE_REGEX.search(path)
        if matches is not None:
            tile.date = matches[0]
         
        # prefixes for opening files with gdal/ogr
        if 's3' in tile.fs.protocol:
            tile.fs_prefix = '/vsis3/'
        else:
            tile.fs_prefix = ''

        # rebuild the path from properties
        # might be overkill
        base_prefix = path.split(f"/{tile.zone}/{tile.quadkey}")[0]
        base_prefix = urljoin(base_prefix, tile.zone, tile.quadkey)
        if tile.date is not None:
            base_prefix = urljoin(base_prefix, tile.date)
        base_prefix = urljoin(base_prefix, tile.acq_id)
        tile.asset_paths = {
            'visual': base_prefix + '-visual.tif',
            'analytic-pan': base_prefix + '-analytic-pan.tif',
            'pan': base_prefix + '-pan.tif',
            'analytic-ms': base_prefix + '-analytic-ms.tif',
            'ms': base_prefix + '-analytic-ms.tif',
            'item': base_prefix + '.json',
            '_clouds': base_prefix + '-clouds.geojson',
            'clouds': base_prefix + '-cloud-mask.gpkg',
            '_extent': base_prefix + '-extent.gpkg',
            'extent': base_prefix + '-data-mask.gpkg'
        }
        tile.base_prefix = base_prefix

        return tile

    def __repr__(self):
        return f'<ARDTile of {self.acq_id} at z{self.zone}-{self.quadkey}>'
    
    def open_asset(self, name):
        asset = self.asset_paths[name]
        if asset.endswith('tif'):
            if not HAS_RASTERIO:
                raise MissingDependency('Rasterio is required to read ARD rasters')
            return rasterio.open(f'{self.fs_prefix}{asset}')

        if asset.endswith('geojson') or asset.endswith('gpkg'): 
            if not HAS_FIONA:
                raise MissingDependency('Fiona is required to access ARD masks as geometries')
            if not HAS_SHAPELY:
                raise MissingDependency('Shapely is required to access ARD masks as geometries')
            with fiona.open(f'{self.fs_prefix}{asset}') as layer:
                geom = [shape(f['geometry']) for f in layer]
                proj = layer.crs['init']
                if proj != 'epsg:4326':
                    tfm = get_transform(proj, 4326)
                    geom = [transform(tfm, f) for f in geom]
            if len(geom) == 1:
                return geom[0]
            elif len(geom) == 0:
                return None
            else:
                return GeometryCollection(geom)
            

        if asset.endswith('json'):
            with self.fs.open(asset) as f:
                return json.load(f)


    def __geo_interface__(self):
        # return cell extents or pixel extents?
        return mapping(self.extent)
